/*
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2011, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage, Inc. nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

/** \author Jia Pan */


#include "fcl/geometric_shapes_utility.h"
#include "fcl/BV_fitter.h"

namespace fcl
{

namespace details
{

std::vector<Vec3f> getBoundVertices(const Box& box, const SimpleTransform& tf)
{
  std::vector<Vec3f> result(8);
  FCL_REAL a = box.side[0] / 2;
  FCL_REAL b = box.side[1] / 2;
  FCL_REAL c = box.side[2] / 2;
  result[0] = tf.transform(Vec3f(a, b, c));
  result[1] = tf.transform(Vec3f(a, b, -c));
  result[2] = tf.transform(Vec3f(a, -b, c));
  result[3] = tf.transform(Vec3f(a, -b, -c));
  result[4] = tf.transform(Vec3f(-a, b, c));
  result[5] = tf.transform(Vec3f(-a, b, -c));
  result[6] = tf.transform(Vec3f(-a, -b, c));
  result[7] = tf.transform(Vec3f(-a, -b, -c));
  
  return result;
}

// we use icosahedron to bound the sphere
std::vector<Vec3f> getBoundVertices(const Sphere& sphere, const SimpleTransform& tf)
{
  std::vector<Vec3f> result(12);
  const FCL_REAL m = (1 + sqrt(5.0)) / 2.0;
  FCL_REAL edge_size = sphere.radius * 6 / (sqrt(27.0) + sqrt(15.0));
  
  FCL_REAL a = edge_size;
  FCL_REAL b = m * edge_size;
  result[0] = tf.transform(Vec3f(0, a, b));
  result[1] = tf.transform(Vec3f(0, -a, b));
  result[2] = tf.transform(Vec3f(0, a, -b));
  result[3] = tf.transform(Vec3f(0, -a, -b));
  result[4] = tf.transform(Vec3f(a, b, 0));
  result[5] = tf.transform(Vec3f(-a, b, 0));
  result[6] = tf.transform(Vec3f(a, -b, 0));
  result[7] = tf.transform(Vec3f(-a, -b, 0));
  result[8] = tf.transform(Vec3f(b, 0, a));
  result[9] = tf.transform(Vec3f(b, 0, -a));
  result[10] = tf.transform(Vec3f(-b, 0, a));
  result[11] = tf.transform(Vec3f(-b, 0, -a));

  return result;
}

std::vector<Vec3f> getBoundVertices(const Capsule& capsule, const SimpleTransform& tf)
{
  std::vector<Vec3f> result(36);
  const FCL_REAL m = (1 + sqrt(5.0)) / 2.0;

  FCL_REAL hl = capsule.lz * 0.5;
  FCL_REAL edge_size = capsule.radius * 6 / (sqrt(27.0) + sqrt(15.0));
  FCL_REAL a = edge_size;
  FCL_REAL b = m * edge_size;
  FCL_REAL r2 = capsule.radius * 2 / sqrt(3.0);


  result[0] = tf.transform(Vec3f(0, a, b + hl));
  result[1] = tf.transform(Vec3f(0, -a, b + hl));
  result[2] = tf.transform(Vec3f(0, a, -b + hl));
  result[3] = tf.transform(Vec3f(0, -a, -b + hl));
  result[4] = tf.transform(Vec3f(a, b, hl));
  result[5] = tf.transform(Vec3f(-a, b, hl));
  result[6] = tf.transform(Vec3f(a, -b, hl));
  result[7] = tf.transform(Vec3f(-a, -b, hl));
  result[8] = tf.transform(Vec3f(b, 0, a + hl));
  result[9] = tf.transform(Vec3f(b, 0, -a + hl));
  result[10] = tf.transform(Vec3f(-b, 0, a + hl));
  result[11] = tf.transform(Vec3f(-b, 0, -a + hl));

  result[12] = tf.transform(Vec3f(0, a, b - hl));
  result[13] = tf.transform(Vec3f(0, -a, b - hl));
  result[14] = tf.transform(Vec3f(0, a, -b - hl));
  result[15] = tf.transform(Vec3f(0, -a, -b - hl));
  result[16] = tf.transform(Vec3f(a, b, -hl));
  result[17] = tf.transform(Vec3f(-a, b, -hl));
  result[18] = tf.transform(Vec3f(a, -b, -hl));
  result[19] = tf.transform(Vec3f(-a, -b, -hl));
  result[20] = tf.transform(Vec3f(b, 0, a - hl));
  result[21] = tf.transform(Vec3f(b, 0, -a - hl));
  result[22] = tf.transform(Vec3f(-b, 0, a - hl));
  result[23] = tf.transform(Vec3f(-b, 0, -a - hl));

  FCL_REAL c = 0.5 * r2;
  FCL_REAL d = capsule.radius;
  result[24] = tf.transform(Vec3f(r2, 0, hl));
  result[25] = tf.transform(Vec3f(c, d, hl));
  result[26] = tf.transform(Vec3f(-c, d, hl));
  result[27] = tf.transform(Vec3f(-r2, 0, hl));
  result[28] = tf.transform(Vec3f(-c, -d, hl));
  result[29] = tf.transform(Vec3f(c, -d, hl));

  result[30] = tf.transform(Vec3f(r2, 0, -hl));
  result[31] = tf.transform(Vec3f(c, d, -hl));
  result[32] = tf.transform(Vec3f(-c, d, -hl));
  result[33] = tf.transform(Vec3f(-r2, 0, -hl));
  result[34] = tf.transform(Vec3f(-c, -d, -hl));
  result[35] = tf.transform(Vec3f(c, -d, -hl));

  return result;
}


std::vector<Vec3f> getBoundVertices(const Cone& cone, const SimpleTransform& tf)
{
  std::vector<Vec3f> result(7);
  
  FCL_REAL hl = cone.lz * 0.5;
  FCL_REAL r2 = cone.radius * 2 / sqrt(3.0);
  FCL_REAL a = 0.5 * r2;
  FCL_REAL b = cone.radius;

  result[0] = tf.transform(Vec3f(r2, 0, -hl));
  result[1] = tf.transform(Vec3f(a, b, -hl));
  result[2] = tf.transform(Vec3f(-a, b, -hl));
  result[3] = tf.transform(Vec3f(-r2, 0, -hl));
  result[4] = tf.transform(Vec3f(-a, -b, -hl));
  result[5] = tf.transform(Vec3f(a, -b, -hl));

  result[6] = tf.transform(Vec3f(0, 0, hl));
                          
  return result;
}

std::vector<Vec3f> getBoundVertices(const Cylinder& cylinder, const SimpleTransform& tf)
{
  std::vector<Vec3f> result(12);

  FCL_REAL hl = cylinder.lz * 0.5;
  FCL_REAL r2 = cylinder.radius * 2 / sqrt(3.0);
  FCL_REAL a = 0.5 * r2;
  FCL_REAL b = cylinder.radius;

  result[0] = tf.transform(Vec3f(r2, 0, -hl));
  result[1] = tf.transform(Vec3f(a, b, -hl));
  result[2] = tf.transform(Vec3f(-a, b, -hl));
  result[3] = tf.transform(Vec3f(-r2, 0, -hl));
  result[4] = tf.transform(Vec3f(-a, -b, -hl));
  result[5] = tf.transform(Vec3f(a, -b, -hl));

  result[6] = tf.transform(Vec3f(r2, 0, hl));
  result[7] = tf.transform(Vec3f(a, b, hl));
  result[8] = tf.transform(Vec3f(-a, b, hl));
  result[9] = tf.transform(Vec3f(-r2, 0, hl));
  result[10] = tf.transform(Vec3f(-a, -b, hl));
  result[11] = tf.transform(Vec3f(a, -b, hl));

  return result;
}

std::vector<Vec3f> getBoundVertices(const Convex& convex, const SimpleTransform& tf)
{
  std::vector<Vec3f> result(convex.num_points);
  for(int i = 0; i < convex.num_points; ++i)
  {
    result[i] = tf.transform(convex.points[i]);
  }

  return result;
}

std::vector<Vec3f> getBoundVertices(const Triangle2& triangle, const SimpleTransform& tf)
{
  std::vector<Vec3f> result(3);
  result[0] = tf.transform(triangle.a);
  result[1] = tf.transform(triangle.b);
  result[2] = tf.transform(triangle.c);

  return result;
}

} // end detail


template<>
void computeBV<AABB, Box>(const Box& s, const SimpleTransform& tf, AABB& bv)
{
  const Matrix3f& R = tf.getRotation();
  const Vec3f& T = tf.getTranslation();

  FCL_REAL x_range = 0.5 * (fabs(R[0][0] * s.side[0]) + fabs(R[0][1] * s.side[1]) + fabs(R[0][2] * s.side[2]));
  FCL_REAL y_range = 0.5 * (fabs(R[1][0] * s.side[0]) + fabs(R[1][1] * s.side[1]) + fabs(R[1][2] * s.side[2]));
  FCL_REAL z_range = 0.5 * (fabs(R[2][0] * s.side[0]) + fabs(R[2][1] * s.side[1]) + fabs(R[2][2] * s.side[2]));

  bv.max_ = T + Vec3f(x_range, y_range, z_range);
  bv.min_ = T + Vec3f(-x_range, -y_range, -z_range);
}

template<>
void computeBV<AABB, Sphere>(const Sphere& s, const SimpleTransform& tf, AABB& bv)
{
  const Vec3f& T = tf.getTranslation();

  bv.max_ = T + Vec3f(s.radius, s.radius, s.radius);
  bv.min_ = T + Vec3f(-s.radius, -s.radius, -s.radius);
}

template<>
void computeBV<AABB, Capsule>(const Capsule& s, const SimpleTransform& tf, AABB& bv)
{
  const Matrix3f& R = tf.getRotation();
  const Vec3f& T = tf.getTranslation();

  FCL_REAL x_range = 0.5 * fabs(R[0][2] * s.lz) + s.radius;
  FCL_REAL y_range = 0.5 * fabs(R[1][2] * s.lz) + s.radius;
  FCL_REAL z_range = 0.5 * fabs(R[2][2] * s.lz) + s.radius;

  bv.max_ = T + Vec3f(x_range, y_range, z_range);
  bv.min_ = T + Vec3f(-x_range, -y_range, -z_range);
}

template<>
void computeBV<AABB, Cone>(const Cone& s, const SimpleTransform& tf, AABB& bv)
{
  const Matrix3f& R = tf.getRotation();
  const Vec3f& T = tf.getTranslation();

  FCL_REAL x_range = fabs(R[0][0] * s.radius) + fabs(R[0][1] * s.radius) + 0.5 * fabs(R[0][2] * s.lz);
  FCL_REAL y_range = fabs(R[1][0] * s.radius) + fabs(R[1][1] * s.radius) + 0.5 * fabs(R[1][2] * s.lz);
  FCL_REAL z_range = fabs(R[2][0] * s.radius) + fabs(R[2][1] * s.radius) + 0.5 * fabs(R[2][2] * s.lz);

  bv.max_ = T + Vec3f(x_range, y_range, z_range);
  bv.min_ = T + Vec3f(-x_range, -y_range, -z_range);
}

template<>
void computeBV<AABB, Cylinder>(const Cylinder& s, const SimpleTransform& tf, AABB& bv)
{
  const Matrix3f& R = tf.getRotation();
  const Vec3f& T = tf.getTranslation();

  FCL_REAL x_range = fabs(R[0][0] * s.radius) + fabs(R[0][1] * s.radius) + 0.5 * fabs(R[0][2] * s.lz);
  FCL_REAL y_range = fabs(R[1][0] * s.radius) + fabs(R[1][1] * s.radius) + 0.5 * fabs(R[1][2] * s.lz);
  FCL_REAL z_range = fabs(R[2][0] * s.radius) + fabs(R[2][1] * s.radius) + 0.5 * fabs(R[2][2] * s.lz);

  bv.max_ = T + Vec3f(x_range, y_range, z_range);
  bv.min_ = T + Vec3f(-x_range, -y_range, -z_range);
}

template<>
void computeBV<AABB, Convex>(const Convex& s, const SimpleTransform& tf, AABB& bv)
{
  const Matrix3f& R = tf.getRotation();
  const Vec3f& T = tf.getTranslation();

  AABB bv_;
  for(int i = 0; i < s.num_points; ++i)
  {
    Vec3f new_p = R * s.points[i] + T;
    bv_ += new_p;
  }

  bv = bv_;
}

template<>
void computeBV<AABB, Triangle2>(const Triangle2& s, const SimpleTransform& tf, AABB& bv)
{
  AABB bv_;
  bv_ += tf.transform(s.a);
  bv_ += tf.transform(s.b);
  bv_ += tf.transform(s.c);

  bv = bv_;
}

template<>
void computeBV<AABB, Plane>(const Plane& s, const SimpleTransform& tf, AABB& bv)
{
  const Matrix3f& R = tf.getRotation();

  Vec3f n = R * s.n;

  AABB bv_;
  if(n[1] == (FCL_REAL)0.0 && n[2] == (FCL_REAL)0.0)
  {
    // normal aligned with x axis
    if(n[0] < 0) bv_.min_[0] = -s.d;
    else if(n[0] > 0) bv_.max_[0] = s.d;
  }
  else if(n[0] == (FCL_REAL)0.0 && n[2] == (FCL_REAL)0.0)
  {
    // normal aligned with y axis
    if(n[1] < 0) bv_.min_[1] = -s.d;
    else if(n[1] > 0) bv_.max_[1] = s.d;
  }
  else if(n[0] == (FCL_REAL)0.0 && n[1] == (FCL_REAL)0.0)
  {
    // normal aligned with z axis
    if(n[2] < 0) bv_.min_[2] = -s.d;
    else if(n[2] > 0) bv_.max_[2] = s.d;
  }

  bv = bv_;
}


template<>
void computeBV<OBB, Box>(const Box& s, const SimpleTransform& tf, OBB& bv)
{
  const Matrix3f& R = tf.getRotation();
  const Vec3f& T = tf.getTranslation();

  bv.To = T;
  bv.axis[0] = R.getColumn(0);
  bv.axis[1] = R.getColumn(1);
  bv.axis[2] = R.getColumn(2);
  bv.extent = s.side * (FCL_REAL)0.5;
}

template<>
void computeBV<OBB, Sphere>(const Sphere& s, const SimpleTransform& tf, OBB& bv)
{
  const Vec3f& T = tf.getTranslation();

  bv.To = T;
  bv.axis[0].setValue(1, 0, 0);
  bv.axis[1].setValue(0, 1, 0);
  bv.axis[2].setValue(0, 0, 1);
  bv.extent.setValue(s.radius);
}

template<>
void computeBV<OBB, Capsule>(const Capsule& s, const SimpleTransform& tf, OBB& bv)
{
  const Matrix3f& R = tf.getRotation();
  const Vec3f& T = tf.getTranslation();

  bv.To = T;
  bv.axis[0] = R.getColumn(0);
  bv.axis[1] = R.getColumn(1);
  bv.axis[2] = R.getColumn(2);
  bv.extent.setValue(s.radius, s.radius, s.lz / 2 + s.radius);
}

template<>
void computeBV<OBB, Cone>(const Cone& s, const SimpleTransform& tf, OBB& bv)
{
  const Matrix3f& R = tf.getRotation();
  const Vec3f& T = tf.getTranslation();

  bv.To = T;
  bv.axis[0] = R.getColumn(0);
  bv.axis[1] = R.getColumn(1);
  bv.axis[2] = R.getColumn(2);
  bv.extent.setValue(s.radius, s.radius, s.lz / 2);
}

template<>
void computeBV<OBB, Cylinder>(const Cylinder& s, const SimpleTransform& tf, OBB& bv)
{
  const Matrix3f& R = tf.getRotation();
  const Vec3f& T = tf.getTranslation();

  bv.To = T;
  bv.axis[0] = R.getColumn(0);
  bv.axis[1] = R.getColumn(1);
  bv.axis[2] = R.getColumn(2);
  bv.extent.setValue(s.radius, s.radius, s.lz / 2);
}

template<>
void computeBV<OBB, Convex>(const Convex& s, const SimpleTransform& tf, OBB& bv)
{
  const Matrix3f& R = tf.getRotation();
  const Vec3f& T = tf.getTranslation();

  fit(s.points, s.num_points, bv);

  bv.axis[0] = R * bv.axis[0];
  bv.axis[1] = R * bv.axis[1];
  bv.axis[2] = R * bv.axis[2];

  bv.To = R * bv.To + T;
}

template<>
void computeBV<OBB, Plane>(const Plane& s, const SimpleTransform& tf, OBB& bv)
{
  const Matrix3f& R = tf.getRotation();
  const Vec3f& T = tf.getTranslation();

  // generate other two axes orthonormal to plane normal
  generateCoordinateSystem(s.n, bv.axis[1], bv.axis[2]);
  bv.axis[0] = s.n;

  bv.extent.setValue(0, std::numeric_limits<FCL_REAL>::max(), std::numeric_limits<FCL_REAL>::max());

  Vec3f p = s.n * s.d;
  bv.To = R * p + T;
}

template<>
void computeBV<RSS, Plane>(const Plane& s, const SimpleTransform& tf, RSS& bv)
{
  const Matrix3f& R = tf.getRotation();
  const Vec3f& T = tf.getTranslation();

  generateCoordinateSystem(s.n, bv.axis[1], bv.axis[2]);
  bv.axis[0] = s.n;

  bv.l[0] = std::numeric_limits<FCL_REAL>::max();
  bv.l[1] = std::numeric_limits<FCL_REAL>::max();

  bv.r = std::numeric_limits<FCL_REAL>::max();
}

template<>
void computeBV<OBBRSS, Plane>(const Plane& s, const SimpleTransform& tf, OBBRSS& bv)
{
}

template<>
void computeBV<kIOS, Plane>(const Plane& s, const SimpleTransform& tf, kIOS& bv)
{
}

template<>
void computeBV<KDOP<16>, Plane>(const Plane& s, const SimpleTransform& tf, KDOP<16>& bv)
{
}

template<>
void computeBV<KDOP<18>, Plane>(const Plane& s, const SimpleTransform& tf, KDOP<18>& bv)
{
}

template<>
void computeBV<KDOP<24>, Plane>(const Plane& s, const SimpleTransform& tf, KDOP<24>& bv)
{
}





}
